@@ -79,8 +79,10 @@
 #include <sys/ioctl.h>
 #include <time.h>
 #include <limits.h>
+#include <strings.h>
 
 #include "compat/compat.h"
+#include "dsp/generated/starch.h"
 
 // ============================= #defines ===============================
 
@@ -91,8 +93,8 @@
 #define MODES_RTL_BUF_SIZE         (16*16384)                 // 256k
 #define MODES_MAG_BUF_SAMPLES      (MODES_RTL_BUF_SIZE / 2)   // Each sample is 2 bytes
 #define MODES_MAG_BUFFERS          12                         // Number of magnitude buffers (should be smaller than RTL_BUFFERS for flowcontrol to work)
-#define MODES_AUTO_GAIN            -100                       // Use automatic gain
-#define MODES_MAX_GAIN             999999                     // Use max available gain
+#define MODES_LEGACY_AUTO_GAIN     -10                        // old gain value for "use automatic gain"
+#define MODES_DEFAULT_GAIN         999999                     // Use default SDR gain
 #define MODES_MSG_SQUELCH_DB       4.0                        // Minimum SNR, in dB
 #define MODES_MSG_ENCODER_ERRS     3                          // Maximum number of encoding errors
 
@@ -163,6 +165,12 @@ typedef enum {
     UNIT_METERS
 } altitude_unit_t;
 
+typedef enum {
+    UNIT_NAUTICAL_MILES,
+    UNIT_STATUTE_MILES,
+    UNIT_KILOMETERS,
+} interactive_distance_unit_t;
+
 typedef enum {
     ALTITUDE_BARO,
     ALTITUDE_GEOM
@@ -195,6 +203,7 @@ typedef enum {
 typedef enum {
     COMMB_UNKNOWN,
     COMMB_AMBIGUOUS,
+    COMMB_NOT_DECODED,
     COMMB_EMPTY_RESPONSE,
     COMMB_DATALINK_CAPS,
     COMMB_GICB_CAPS,
@@ -202,7 +211,9 @@ typedef enum {
     COMMB_ACAS_RA,
     COMMB_VERTICAL_INTENT,
     COMMB_TRACK_TURN,
-    COMMB_HEADING_SPEED
+    COMMB_HEADING_SPEED,
+    COMMB_MRAR,
+    COMMB_AIRBORNE_POSITION
 } commb_format_t;
 
 typedef enum {
@@ -230,6 +241,24 @@ typedef enum {
     NAV_ALT_INVALID, NAV_ALT_UNKNOWN, NAV_ALT_AIRCRAFT, NAV_ALT_MCP, NAV_ALT_FMS
 } nav_altitude_source_t;
 
+// BDS4,4 MRAR - FOM/Source values
+typedef enum {
+   MRAR_SOURCE_INVALID = 0,
+   MRAR_SOURCE_INS = 1,
+   MRAR_SOURCE_GNSS = 2,
+   MRAR_SOURCE_DMEDME = 3,
+   MRAR_SOURCE_VORDME = 4,
+   MRAR_SOURCE_RESERVED = 5
+} mrar_source_t;
+
+// BDS4,4 and BDS4,5 hazard reports
+typedef enum {
+   HAZARD_NIL = 0,
+   HAZARD_LIGHT = 1,
+   HAZARD_MODERATE = 2,
+   HAZARD_SEVERE = 3
+} hazard_t;
+
 #define MODES_NON_ICAO_ADDRESS       (1<<24) // Set on addresses to indicate they are not ICAO addresses
 
 #define MODES_INTERACTIVE_REFRESH_TIME 250      // Milliseconds
@@ -249,6 +278,9 @@ typedef enum {
 #define MAX_AMPLITUDE 65535.0
 #define MAX_POWER (MAX_AMPLITUDE * MAX_AMPLITUDE)
 
+#define FAUP_DEFAULT_RATE_MULTIPLIER    1.0                  // FA Upload rate multiplier
+
+
 // Include subheaders after all the #defines are in place
 
 #include "util.h"
@@ -262,11 +294,12 @@ typedef enum {
 #include "convert.h"
 #include "sdr.h"
 #include "fifo.h"
+#include "adaptive.h"
 
 //======================== structure declarations =========================
 
 typedef enum {
-    SDR_NONE, SDR_IFILE, SDR_RTLSDR, SDR_BLADERF, SDR_HACKRF, SDR_LIMESDR
+    SDR_NONE, SDR_IFILE, SDR_RTLSDR, SDR_BLADERF, SDR_HACKRF, SDR_LIMESDR, SDR_SOAPYSDR
 } sdr_type_t;
 
 // Program global state
@@ -286,9 +319,9 @@ struct _Modes {                             // Internal state
     // Sample conversion
     int            dc_filter;        // should we apply a DC filter?
 
-    // RTLSDR
+    // RTLSDR and some other SDRs
     char *        dev_name;
-    int           gain;
+    float         gain;              // value in dB, or MODES_AUTO_GAIN, or MODES_MAX_GAIN
     int           freq;
 
     // Networking
@@ -296,11 +329,13 @@ struct _Modes {                             // Internal state
     struct net_service *services;    // Active services
     struct client *clients;          // Our clients
 
-    struct net_service *beast_verbatim_service;  // Beast-format output service, verbatim mode
-    struct net_service *beast_cooked_service;    // Beast-format output service, "cooked" mode
+    struct net_service *beast_verbatim_service;        // Beast-format output service, verbatim mode
+    struct net_service *beast_verbatim_local_service;  // Beast-format output service, verbatim+local mode
+    struct net_service *beast_cooked_service;          // Beast-format output service, "cooked" mode
 
     struct net_writer raw_out;                   // AVR-format output
     struct net_writer beast_verbatim_out;        // Beast-format output, verbatim mode
+    struct net_writer beast_verbatim_local_out;  // Beast-format output, verbatim+local mode
     struct net_writer beast_cooked_out;          // Beast-format output, "cooked" mode
     struct net_writer sbs_out;                   // SBS-format output
     struct net_writer stratux_out;               // Stratux-format output
@@ -314,6 +349,8 @@ struct _Modes {                             // Internal state
     sdr_type_t sdr_type;             // where are we getting data from?
     int   nfix_crc;                  // Number of crc bit error(s) to correct
     int   check_crc;                 // Only display messages with good CRC
+    int   fix_df;                    // Try to correct damage to the DF field, as well as the main message body
+    int   enable_df24;               // Enable decoding of DF24..DF31 (Comm-D ELM)
     int   raw;                       // Raw output format
     int   mode_ac;                   // Enable decoding of SSR Modes A & C
     int   mode_ac_auto;              // allow toggling of A/C by Beast commands
@@ -336,6 +373,10 @@ struct _Modes {                             // Internal state
     uint32_t show_only;              // Only show messages from this ICAO
     int   interactive;               // Interactive mode
     uint64_t interactive_display_ttl;// Interactive mode: TTL display
+    int interactive_display_size;    // Size of TTL display
+    int   interactive_show_distance; // Show aircraft distance and bearing instead of lat/lon
+    interactive_distance_unit_t interactive_distance_units; // Units for interactive distance display
+    char *interactive_callsign_filter; // Filter for interactive display callsigns
     uint64_t stats;                  // Interval (millis) between stats dumps,
     int   stats_range_histo;         // Collect/show a range histogram?
     int   onlyaddr;                  // Print only ICAO addresses
@@ -344,7 +385,10 @@ struct _Modes {                             // Internal state
     int   mlat;                      // Use Beast ascii format for raw data output, i.e. @...; iso *...;
     char *json_dir;                  // Path to json base directory, or NULL not to write json.
     uint64_t json_interval;          // Interval between rewriting the json aircraft file, in milliseconds; also the advertised map refresh interval
+    uint64_t json_stats_interval;    // Interval between rewriting the json stats file, in milliseconds
     int   json_location_accuracy;    // Accuracy of location metadata: 0=none, 1=approx, 2=exact
+    double faup_rate_multiplier;     // Multiplier to adjust rate of faup1090 messages emitted
+    bool faup_upload_unknown_commb;  // faup1090: should we upload Comm-B messages that weren't in a recognized format?
 
     int   json_aircraft_history_next;
     struct {
@@ -362,13 +406,36 @@ struct _Modes {                             // Internal state
     struct aircraft *aircrafts;
 
     // Statistics
-    struct stats stats_current;
-    struct stats stats_alltime;
-    struct stats stats_periodic;
-    struct stats stats_1min[15];
-    int stats_latest_1min;
-    struct stats stats_5min;
-    struct stats stats_15min;
+    struct stats stats_current;     // Currently accumulating stats, this is where all stats are initially collected
+    struct stats stats_alltime;     // Accumulated stats since the start of the process
+    struct stats stats_periodic;    // Accumulated stats since the last periodic stats display (--stats-every)
+    struct stats stats_latest;      // Accumulated stats since the end of the last 1-minute period
+    struct stats stats_1min[15];    // Accumulated stats for a full 1-minute window; this is a ring buffer maintaining a history of 15 minutes
+    int stats_newest_1min;          // Index into stats_1min of the most recent 1-minute window
+    struct stats stats_5min;        // Accumulated stats from the last 5 complete 1-minute windows
+    struct stats stats_15min;       // Accumulated stats from the last 15 complete 1-minute windows
+
+    // Adaptive gain config
+    float adaptive_min_gain_db;
+    float adaptive_max_gain_db;
+
+    float adaptive_duty_cycle;
+
+    bool adaptive_burst_control;
+    float adaptive_burst_alpha;
+    unsigned adaptive_burst_change_delay;
+    float adaptive_burst_loud_rate;
+    unsigned adaptive_burst_loud_runlength;
+    float adaptive_burst_quiet_rate;
+    unsigned adaptive_burst_quiet_runlength;
+
+    bool adaptive_range_control;
+    float adaptive_range_alpha;
+    unsigned adaptive_range_percentile;
+    float adaptive_range_target;
+    unsigned adaptive_range_change_delay;
+    unsigned adaptive_range_scan_delay;
+    unsigned adaptive_range_rescan_delay;
 };
 
 extern struct _Modes Modes;
@@ -511,21 +578,20 @@ struct modesMessage {
         unsigned nic_c : 1;        // if nic_c_valid
         unsigned nic_baro : 1;     // if nic_baro_valid
 
-        unsigned nac_p : 4;        // if nac_p_valid
-        unsigned nac_v : 3;        // if nac_v_valid
+        unsigned nac_p;        // if nac_p_valid
+        unsigned nac_v;        // if nac_v_valid
 
-        unsigned sil : 2;          // if sil_type != SIL_INVALID
+        unsigned sil;          // if sil_type != SIL_INVALID
         sil_type_t sil_type;
 
-        unsigned gva : 2;          // if gva_valid
-
-        unsigned sda : 2;          // if sda_valid
+        unsigned gva;          // if gva_valid
+        unsigned sda;          // if sda_valid
     } accuracy;
 
     // Operational Status
     struct {
         unsigned valid : 1;
-        unsigned version : 3;
+        unsigned version;
 
         unsigned om_acas_ra : 1;
         unsigned om_ident : 1;
@@ -537,7 +603,7 @@ struct modesMessage {
         unsigned cc_1090_in : 1;
         unsigned cc_arv : 1;
         unsigned cc_ts : 1;
-        unsigned cc_tc : 2;
+        unsigned cc_tc;
         unsigned cc_uat_in : 1;
         unsigned cc_poa : 1;
         unsigned cc_b2_low : 1;
@@ -562,14 +628,30 @@ struct modesMessage {
 
         float    heading;       // heading, degrees (0-359) (could be magnetic or true heading; magnetic recommended)
         heading_type_t heading_type;
-        unsigned fms_altitude;  // FMS selected altitude
-        unsigned mcp_altitude;  // MCP/FCU selected altitude
+        int      fms_altitude;  // FMS selected altitude
+        int      mcp_altitude;  // MCP/FCU selected altitude
         float    qnh;           // altimeter setting (QFE or QNH/QNE), millibars
 
         nav_altitude_source_t altitude_source;
 
         nav_modes_t modes;
     } nav;
+
+    // BDS 4,4 MRAR
+    unsigned mrar_source_valid : 1;
+    unsigned wind_valid : 1;
+    unsigned temperature_valid : 1;
+    unsigned pressure_valid : 1;
+    unsigned turbulence_valid : 1;
+    unsigned humidity_valid : 1;
+
+    mrar_source_t mrar_source;
+    float wind_speed;    // kts
+    float wind_dir;      // degrees
+    float temperature;   // degrees C
+    float pressure;      // hPa
+    hazard_t turbulence; // NIL/LIGHT/MODERATE/SEVERE
+    float humidity;      // 0-100 %
 };
 
 // This one needs modesMessage:
