@@ -25,7 +25,12 @@ void rtlsdrInitConfig();
 void rtlsdrShowHelp();
 bool rtlsdrOpen();
 void rtlsdrRun();
+void rtlsdrStop();
 void rtlsdrClose();
 bool rtlsdrHandleOption(int argc, char **argv, int *jptr);
+int rtlsdrGetGain();
+int rtlsdrGetMaxGain();
+double rtlsdrGetGainDb(int step);
+int rtlsdrSetGain(int step);
 
 #endif
